//----------------------------------------------------------------------------
// DOSSaver - Screen Saver p/ DOS
// by Wenderson Teixeira
//
// Este programa e fornecido com codigo fonte para motivos de aprendizado, 
// e pode ser alterado, recompilado, etc, da forma que se achar necessaria
// contanto que se mantenha o codigo fonte gratis.
// Sua venda e proibida, exceto em casos em que se tenha uma permissao 
// explicita do autor para tal.
//
// Wenderson Teixeira
//----------------------------------------------------------------------------

#include <dos.h>
#include <stdio.h>
#include <conio.h>
#include <alloc.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

// Para debugar passo a passo, retire o comentario da linha abaixo
//#define DEBUG

#ifdef DEBUG
#include "readkey.h"
#endif

// Em C, utiliza-se 'void interrupt (*)(void)' em C++, 'void interrupt(*)(...)'
#ifdef __cplusplus
#define PARAMS  ...
#else
#define PARAMS void
#endif

// Definicao das interrupcoes utilizadas
#define TIMERINTR     0x1C
#define KEYBOARDINTR  0x09

// Definicao de constantes
#define false 0
#define true  !false

// Definicao de tipos
typedef int bool;

// Variaveis utilizadas pelo ScreenSaver
bool IsSaverActive;
int Minutes, Seconds, TimeToSave;
char SavedScreen[4000];
struct text_info TextInfo;
void (*SaverFunction)();

// Ponteiros p/ os controladores de interrupcao antigos
void interrupt (*OldTimerHandler)(PARAMS);
void interrupt (*OldKeyboardHandler)(PARAMS);

//-------------------------------------------------------------------
// Definicao dos novos controladores de interrupcao
//-------------------------------------------------------------------

// Controlador do Timer ---------------------------------------------
#ifdef DEBUG
void TimerHandler()
#else
void interrupt TimerHandler(PARAMS)
#endif
{
  static int counter = 0;

  // Se o screensaver nao estiver ativo, realiza contagem
  if(!IsSaverActive)
  {
    // Se ja esta na hora, aciona o screensaver
    if((Minutes * 60 + Seconds) >= TimeToSave)
    {
      IsSaverActive = true;
      gettextinfo(&TextInfo);
      _setcursortype(_NOCURSOR);
      gettext(1, 1, TextInfo.screenwidth, TextInfo.screenheight, SavedScreen);
    }
    else
    {
      // Contagem do tempo
      (++counter % 18) == 0 ? Seconds++ : 0;
      (counter == 1092) ? (counter = 0, Seconds = 0, Minutes++) : 0;
    }

    /*
    // Deve-se usar cprintf ou outro tipo de rotina que
    // imprima direta p/ a memoria de video, pois printf
    // chama as rotinas do DOS p/ imprimir, gerando
    // reentrancia, fazendo com que o programa trave
    gotoxy(76, 1);
    cprintf("%02d:%02d", Minutes, Seconds);
    */
  }
  else
  {
    // Se o screensaver estiver ativo e o tempo for ressetado
    // o screensaver deve ser parado
    if(!Minutes && !Seconds)
    {
      textmode(TextInfo.currmode);
      textattr(TextInfo.attribute);
      puttext(1, 1, TextInfo.screenwidth, TextInfo.screenheight, SavedScreen);
      gotoxy(TextInfo.curx, TextInfo.cury);
      _setcursortype(_NORMALCURSOR);
      IsSaverActive = false;
    }
    else
      // Executa screensaver
      SaverFunction();
  }


  #ifndef DEBUG
  // Chama o handler antigo do timer
  OldTimerHandler();
  #endif
}
//-------------------------------------------------------------------

// Controlador do Teclado -------------------------------------------
#ifdef DEBUG
void KeyboardHandler()
#else
void interrupt KeyboardHandler(PARAMS)
#endif
{
  // Se alguma tecla for pressionada, zera o tempo
  Minutes = Seconds = 0;

  #ifndef DEBUG
  // Chama o handler antigo do teclado
  OldKeyboardHandler();
  #endif
}
//-------------------------------------------------------------------

//-------------------------------------------------------------------
// OBS.: Atualmente estas duas funcoes nao estao sendo chamadas, pois
// estou utilizando apenas modo texto, e o Borland C++ possui funcoes
// para isso
int GetVideoMode()
{
  union REGS r;

  // Pega o modo de video corrente
  r.h.ah = 15;
  int86(0x10, &r, &r);

  return r.h.al;
}
//-------------------------------------------------------------------

//-------------------------------------------------------------------
void SetVideoMode(int mode)
{
  union REGS r;

  r.h.ah = 10;
  r.h.al = mode;
  int86(0x10, &r, &r);
}
//-------------------------------------------------------------------

//-------------------------------------------------------------------
// O codigo fonte p/ o screen saver deve ser posto aqui, pois esta 
// funcao sera chamada +/- 18,2 vezes por segundo, quando o Screen
// Saver for acionado.
// Evite parar dentro desta funcao, pois ela sera chamada de dentro 
// de uma interrupcao repetitivamente.
//
// Providencia a animacao
void Saver()
{
  static int x = random(TextInfo.screenwidth), y = 1, c = 1 + random(15);

  // Apaga anterior
  gotoxy(x + 1, y - 1);
  textcolor(TextInfo.normattr);
  putch(' ');

  // Desenha novo
  gotoxy(x--, y++);
  textcolor(c);
  putch('*');

  // reseta coordenadas
  if((y >= 25) || (x <= 0))
  {
    gotoxy(x + 1, y - 1);
    textcolor(TextInfo.normattr);
    putch(' ');

    x = random(TextInfo.screenwidth);
    y = 1;
    c = 1 + random(15);
  }
}
//-------------------------------------------------------------------

//-------------------------------------------------------------------
void InitSaver(int seconds)
{
  // Assegura-se que E/S sera feita direta p/ o video
  _directvideo = 1;

  // Inicializa as variaveis necessarias ao ScreenSaver
  IsSaverActive = false;
  Minutes = Seconds = 0;
  TimeToSave = seconds;
  memset(SavedScreen, 0, sizeof(SavedScreen));
  SaverFunction = Saver;
  gettextinfo(&TextInfo);
  randomize();

  #ifndef DEBUG
  // Seta vetores de interrupcao
  OldTimerHandler = getvect(TIMERINTR);
  OldKeyboardHandler = getvect(KEYBOARDINTR);

  setvect(TIMERINTR, TimerHandler);
  setvect(KEYBOARDINTR, KeyboardHandler);
  #endif
}
//-------------------------------------------------------------------

//-------------------------------------------------------------------
void FinishSaver()
{
  #ifndef DEBUG
  setvect(TIMERINTR, OldTimerHandler);
  setvect(KEYBOARDINTR, OldKeyboardHandler);
  #endif
}
//-------------------------------------------------------------------

#ifdef DEBUG
// Esta funcao tem o objetivo de simular o comportamento do
// screensaver p/ podermos debuggar
void Run()
{
  int key;
  do
  {
    key = ReadKeyNoWait();
    if(key)
      KeyboardHandler();

    TimerHandler();
    delay(50);
  } while(!key);
}
#endif

//-------------------------------------------------------------------
int main(int ArgC, char *ArgV[])
{
  printf("DOSSaver  Copyright VIC 1998\n"
         "\tBy Wenderson Teixeira\n");

  if(ArgC < 2)
    InitSaver(60);
  else
  {
    if(tolower(ArgV[1][1]) == 'h')
    {
      printf("\nDOSSaver e um ScreenSaver p/ DOS em modo texto\n"
             "\tSintaxe: dossaver [options]\n"
             "\tOnde options:\n"
             "\t\t-h - esta tela de help\n"
             "\t\tn  - tempo de espera em segundos\n");
      return 0;
    }
    else
      InitSaver(atoi(ArgV[1]));
  }

  printf("\n\tAguarde %d segundos para ver o ScreenSaver\n", TimeToSave);

#ifndef DEBUG
  // Ao se alterar o programa, deve-se se certificar de que o quantidade
  // de paragrafos seja alterada para comportar as modificacoes.
  // Cada paragrafo e equivalente a 16bytes, entao para se calcular
  // o tamanho de paragrafos deve-se dividir o tamanho do executavel
  // por 16 e arrendondar p/ cima, e recomendavel manter-se a funcao
  // main sempre no final do arquivo

  // Termina e mantem residente
  keep(0, 4000);
#else
  Run();
  FinishSaver();
#endif

  return 0;
}
//-------------------------------------------------------------------

